import imghdr
import os
from typing import List

import supervisely.convert.pointcloud.sly.sly_pointcloud_helper as helpers
from supervisely import PointcloudAnnotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.io.fs import JUNK_FILES, get_file_ext, get_file_name
from supervisely.io.json import load_json_file
from supervisely.pointcloud.pointcloud import validate_ext as validate_pcd_ext


class SLYPointcloudConverter(PointcloudConverter):

    def __str__(self) -> str:
        return AvailablePointcloudConverters.SLY

    @property
    def ann_ext(self) -> str:
        return ".json"

    @property
    def key_file_ext(self) -> str:
        return ".json"

    def generate_meta_from_annotation(self, ann_path: str, meta: ProjectMeta) -> ProjectMeta:
        meta = helpers.get_meta_from_annotation(ann_path, meta)
        return meta

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        try:
            ann_json = load_json_file(ann_path)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            ann = PointcloudAnnotation.from_json(ann_json, meta)
            return True
        except Exception as e:
            return False

    def validate_key_file(self, key_file_path: str) -> bool:
        try:
            self._meta = ProjectMeta.from_json(load_json_file(key_file_path))
            return True
        except Exception:
            return False

    def validate_format(self) -> bool:
        detected_ann_cnt = 0
        pcd_list, ann_dict, rimg_dict, rimg_ann_dict = [], {}, {}, {}
        used_img_ext = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                if file == "key_id_map.json":
                    continue
                if file == "meta.json":
                    is_valid = self.validate_key_file(full_path)
                    if is_valid:
                        continue

                ext = get_file_ext(full_path)
                if file in JUNK_FILES:  # add better check
                    continue
                elif ext in self.ann_ext:
                    dir_name = os.path.basename(root)
                    parent_dir_name = os.path.basename(os.path.dirname(root))
                    if any(
                        p.replace("_", " ") in ["images", "related images", "photo context"]
                        for p in [dir_name, parent_dir_name]
                    ) or dir_name.endswith("_pcd"):
                        rimg_ann_dict[file] = full_path
                    else:
                        ann_dict[file] = full_path
                elif imghdr.what(full_path):
                    rimg_dict[file] = full_path
                    if ext not in used_img_ext:
                        used_img_ext.append(ext)
                else:
                    try:
                        validate_pcd_ext(ext)
                        pcd_list.append(full_path)
                    except:
                        continue

        if self._meta is not None:
            meta = self._meta
        else:
            meta = ProjectMeta()

        # create Items
        self._items = []
        for pcd_path in pcd_list:
            name_noext = get_file_name(pcd_path)
            ann_or_rimg_detected = False
            item = self.Item(pcd_path)
            ann_name = f"{item.name}.json"
            if ann_name not in ann_dict:
                ann_name = f"{name_noext}.json"
            if ann_name in ann_dict:
                ann_path = ann_dict[ann_name]
                if self._meta is None:
                    meta = self.generate_meta_from_annotation(ann_path, meta)
                is_valid = self.validate_ann_file(ann_path, meta)
                if is_valid:
                    item.ann_data = ann_path
                    ann_or_rimg_detected = True
            rimg, rimg_ann = helpers.find_related_items(
                item.name, used_img_ext, rimg_dict, rimg_ann_dict
            )
            if rimg is not None and rimg_ann is not None:
                item.set_related_images((rimg, rimg_ann))
                ann_or_rimg_detected = True

            if ann_or_rimg_detected:
                detected_ann_cnt += 1
            self._items.append(item)
        self._meta = meta
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: PointcloudConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> PointcloudAnnotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        if item.ann_data is None:
            return item.create_empty_annotation()

        try:
            ann_json = load_json_file(item.ann_data)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            if renamed_classes or renamed_tags:
                ann_json = helpers.rename_in_json(ann_json, renamed_classes, renamed_tags)
            return PointcloudAnnotation.from_json(ann_json, meta)
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return item.create_empty_annotation()



# coding: utf-8

from supervisely.collection.str_enum import StrEnum


class ProjectType(StrEnum):
    IMAGES = "images"
    VIDEOS = "videos"
    VOLUMES = "volumes"
    POINT_CLOUDS = "point_clouds"
    POINT_CLOUD_EPISODES = "point_cloud_episodes"



# Constants for multispectral and multiview projects.
_MULTISPECTRAL_TAG_NAME = "multispectral"
_MULTIVIEW_TAG_NAME = "multiview"



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class PerClassOutcomeCounts(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.clickable: bool = True
        self.switchable: bool = True
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_class_outcome_counts_1=Widget.Markdown(
                title="Outcome Counts by Class",
                is_header=True,
                formats=[
                    self._loader.vis_texts.definitions.true_positives,
                    self._loader.vis_texts.definitions.false_positives,
                    self._loader.vis_texts.definitions.false_negatives,
                ],
            ),
            markdown_class_outcome_counts_2=Widget.Markdown(
                formats=[self._loader.vis_texts.definitions.f1_score]
            ),
            collapse_perclass_outcome=Widget.Collapse(
                Schema(
                    self._loader.vis_texts,
                    markdown_normalization=Widget.Markdown(title="Normalization"),
                )
            ),
            chart_01=Widget.Chart(switch_key="normalized"),
            chart_02=Widget.Chart(switch_key="absolute"),
        )

    def get_figure(self, widget: Widget.Chart):  #  -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error

        # Per-class Counts
        iou_thres = 0

        tp = self._loader.mp.true_positives[:, iou_thres]
        fp = self._loader.mp.false_positives[:, iou_thres]
        fn = self._loader.mp.false_negatives[:, iou_thres]

        # normalize
        support = tp + fn
        with np.errstate(invalid="ignore", divide="ignore"):
            tp_rel = tp / support
            fp_rel = fp / support
            fn_rel = fn / support

            # sort by f1
            sort_scores = 2 * tp / (2 * tp + fp + fn)

        K = len(self._loader.mp.cat_names)
        sort_indices = np.argsort(sort_scores)
        cat_names_sorted = [self._loader.mp.cat_names[i] for i in sort_indices]
        tp_rel, fn_rel, fp_rel = tp_rel[sort_indices], fn_rel[sort_indices], fp_rel[sort_indices]

        objects_count = np.concatenate([tp[sort_indices], fn[sort_indices], fp[sort_indices]])
        data = {
            "Type": ["TP"] * K + ["FN"] * K + ["FP"] * K,
            "category": cat_names_sorted * 3,
        }
        y_label = ""
        if widget.switch_key == "normalized":
            y_label = "Objects Fraction"
            # Stacked per-class counts
            data["count"] = np.concatenate([tp_rel, fn_rel, fp_rel])
        elif widget.switch_key == "absolute":
            y_label = "Objects Count"
            data["count"] = objects_count

        df = pd.DataFrame(data)

        color_map = {"TP": "#8ACAA1", "FN": "#dd3f3f", "FP": "#F7ADAA"}
        fig = px.bar(
            df,
            x="category",
            y="count",
            color="Type",
            # title="Per-class Outcome Counts",
            height=500,
            width=1000,
            labels={"count": y_label, "category": "Class"},
            color_discrete_map=color_map,
        )
        xaxis_title = fig.layout.xaxis.title.text
        yaxis_title = fig.layout.yaxis.title.text
        if widget.switch_key == "normalized":

            fig.update_traces(
                hovertemplate="Type=%{fullData.name} <br>"
                + xaxis_title
                + "=%{x}<br>"
                + yaxis_title
                + "=%{y:.2f}<extra></extra>"
                # "Images count=%{y:.2f}<extra></extra>"
            )
        elif widget.switch_key == "absolute":
            fig.update_traces(
                hovertemplate="Type=%{fullData.name} <br>"
                + xaxis_title
                + "=%{x}<br>"
                + yaxis_title
                + "=%{y}<extra></extra>",
            )
        return fig

    def get_click_data(self, widget: Widget.Chart) -> Optional[dict]:
        if not self.clickable:
            return
        res = {}
        res["layoutTemplate"] = [None, None, None]

        res["clickData"] = {}
        for class_name, v1 in self._loader.click_data.outcome_counts_by_class.items():
            for outcome, matches_data in v1.items():
                key = class_name + self._keypair_sep + outcome
                res["clickData"][key] = {}
                res["clickData"][key]["imagesIds"] = []
                res["clickData"][key][
                    "title"
                ] = f"Images with objects of class '{class_name}' and outcome '{outcome}'"

                img_ids = set()
                for match_data in matches_data:
                    img_comparison_data = self._loader.comparison_data[match_data["gt_img_id"]]
                    if outcome == "FN":
                        img_ids.add(img_comparison_data.diff_image_info.id)
                    else:
                        img_ids.add(img_comparison_data.pred_image_info.id)
                res["clickData"][key]["imagesIds"] = list(img_ids)
                res["clickData"][key]["filters"] = [
                    {"type": "tag", "tagId": "confidence", "value": [0, 1]},
                    {"type": "tag", "tagId": "outcome", "value": outcome},
                ]
        return res



from dataclasses import dataclass


class RuntimeType:
    PYTORCH = "PyTorch"
    ONNXRUNTIME = "ONNXRuntime"
    TENSORRT = "TensorRT"


class ModelSource:
    PRETRAINED = "Pretrained models"
    CUSTOM = "Custom models"


@dataclass
class CheckpointInfo:
    checkpoint_name: str = None     # e.g. "YOLOv8-L (COCO)"
    architecture: str = None        # e.g. "YOLOv8"
    model_source: str = None        # e.g. "Pretrained models"
    checkpoint_url: str = None
    custom_checkpoint_path: str = None


@dataclass
class DeployInfo:
    checkpoint_name: str
    architecture: str
    model_source: str
    checkpoint_url: str
    custom_checkpoint_path: str
    task_type: str
    device: str
    runtime: str
    hardware: str
    deploy_params: dict


def get_hardware_info(idx: int = 0) -> str:
    import subprocess
    try:
        gpus = subprocess.check_output(["nvidia-smi", "--query-gpu=name", "--format=csv,noheader,nounits"]).decode("utf-8").strip()
        gpu_list = gpus.split("\n")
        if idx >= len(gpu_list):
            raise ValueError(f"No GPU found at index {idx}")
        return gpu_list[idx]
    except subprocess.CalledProcessError:
        return "CPU"


from __future__ import annotations

from typing import TYPE_CHECKING, List

import pandas as pd

from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class ClasswiseErrorAnalysis(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.cv_tasks: List[CVTask] = [CVTask.SEMANTIC_SEGMENTATION.value]

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        import plotly.graph_objects as go  # pylint: disable=import-error

        pd.options.mode.chained_assignment = None  # TODO rm later

        df = self._loader.result_df
        df.drop(["mean"], inplace=True)
        df = df[["IoU", "E_extent_oU", "E_boundary_oU", "E_segment_oU"]]
        df.sort_values(by="IoU", ascending=False, inplace=True)
        labels = list(df.index)
        color_palette = ["cornflowerblue", "moccasin", "lightgreen", "orangered"]

        fig = go.Figure()
        for i, column in enumerate(df.columns):
            fig.add_trace(
                go.Bar(
                    name=column,
                    y=df[column],
                    x=labels,
                    marker_color=color_palette[i],
                )
            )
        fig.update_yaxes(range=[0, 1])
        fig.update_layout(
            barmode="stack",
            plot_bgcolor="rgba(0, 0, 0, 0)",
            title={
                "text": "Classwise segmentation error analysis",
                "y": 0.9,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
            },
        )
        return fig



from typing import List


def generate_main_template(metric_visualizations: List):
    template_str = """<div>
    <sly-iw-sidebar :options="{ height: 'calc(100vh - 130px)', clearMainPanelPaddings: true, leftSided: false,  disableResize: true, sidebarWidth: 300 }">
        <div slot="sidebar">"""

    for vis in metric_visualizations:
        template_str += vis.template_sidebar_str

    template_str += """\n        </div>
      
        <div style="padding: 0 15px;">"""

    for vis in metric_visualizations:
        template_str += """\n                <div style="margin-top: 20px;">"""
        template_str += vis.template_main_str
        template_str += """\n                </div>"""

    template_str += "\n        </div>\n    </sly-iw-sidebar>"

    template_str += """\n
        <sly-iw-gallery
            ref='modal_general'
            iw-widget-id='modal_general'
            :options="{'isModalWindow': true}"
            :actions="{
                'init': {
                'dataSource': '/data/modal_general.json',
                },
            }"
            :command="command"
            :data="data"
        /> \n
        <sly-iw-gallery
            ref='modal_general_diff'
            iw-widget-id='modal_general_diff'
            :options="{'isModalWindow': true}"
            :actions="{
                'init': {
                'dataSource': '/data/modal_general_diff.json',
                },
                'chart-click': {
                    'dataSource': '/data/gallery_explorer_grid_diff_data.json',
                    'galleryId': 'modal_general',
                        'limit': 3,
                },
            }"
            :command="command"
            :data="data"
        > \n
            <span slot="image-left-header">
                <i class="zmdi zmdi-collection-image"></i> Compare with GT
            </span>
        </sly-iw-gallery>
        </div>"""

    return template_str


template_markdown_str = """
            <sly-iw-markdown
              id="{{ widget_id }}"
              class="markdown-no-border"
              iw-widget-id="{{ widget_id }}"
              :actions="{
                'init': {
                  'dataSource': '{{ data_source }}',
                },
              }"
              :command="{{ command }}"
              :data="{{ data }}"
            />
"""

template_chart_str = """
            <sly-iw-chart
              iw-widget-id="{{ widget_id }}"{% if switchable %}
              v-show="state.{{ radio_group }} === '{{ switch_key }}'"
              {% endif %}:actions="{
                'init': {
                  'dataSource': '{{ init_data_source }}',
                },{% if chart_click_data_source %}
                'chart-click': {
                  'dataSource': '{{ chart_click_data_source }}',{% if cls_name in ['outcome_counts'] %}
                  'getKey': (payload) => payload.points[0].data.name,{% endif %}{% if cls_name in ['frequently_confused', 'recall', 'precision', 'recall_vs_precision'] %}
                  'getKey': (payload) => payload.points[0].label,{% endif %}{% if cls_name in ['pr_curve_by_class'] %}
                  'getKey': (payload) => payload.points[0].data.legendgroup,{% endif %}{% if cls_name in ['per_class_avg_precision'] %}
                  'getKey': (payload) => payload.points[0].theta,{% endif %}{% if cls_name in ['per_class_outcome_counts'] %}
                  'getKey': (payload) => `${payload.points[0].label}${'-'}${payload.points[0].data.name}`,{% endif %}{% if cls_name in ['confusion_matrix', 'per_class_outcome_counts'] %}
                  'keySeparator': '{{ key_separator }}',{% endif %}
                  'galleryId': 'modal_general',
                  'limit': 9
                },{% endif %}
              }"
              :command="{{ command }}"
              :data="{{ data }}"
            />


"""

template_radiogroup_str = """<el-radio v-model="state.{{ radio_group }}" label="{{ switch_key }}" style="margin-top: 10px;">{{ switch_key }}</el-radio>"""


template_gallery_str = """<sly-iw-gallery
              iw-widget-id="{{ widget_id }}"
              {% if is_table_gallery %}
              ref='{{ widget_id }}'
              {% endif %}
              :actions="{
                'init': {
                  'dataSource': '{{ init_data_source }}',
                },
                {% if gallery_diff_data_source %}
                'chart-click': {
                    'dataSource': '{{ gallery_diff_data_source }}',
                    'getKey':(payload)=>payload['annotation']['image_id'],
                    'galleryId': 'modal_general',
                        'limit': 3,
                },
                {% endif %}
              }"
              :command="{{ command }}"
              :data="{{ data }}"
            >
                {% if gallery_diff_data_source %}
                    <span slot="image-left-header">
                        <i class="zmdi zmdi-collection-image"></i> Compare with GT
                    </span>
                {% endif %}
            </sly-iw-gallery>

              {% if gallery_click_data_source %}
              <div style="display: flex; justify-content: center; margin-top:10px;" >                
                <el-button iw-widget-id="btn-1" type="primary" @click="command({
                  method: 'update-gallery',
                  payload: {
                    data: {
                      'key': 'explore',
                      'limit': 9,
                      'dataSource': '{{ gallery_click_data_source }}',
                    },
                    'galleryId': 'modal_general_diff',
                  },
                  internalCommand: true
                })">Explore all predictions</el-button>
                </div> {% endif %}
"""


template_table_str = """<sly-iw-table
                iw-widget-id="{{ widget_id }}"
                style="cursor: pointer;"
                :options="{ isRowClickable: true }" 
                :actions="{
                  'init': {
                    'dataSource': '{{ init_data_source }}',
                  },
                  'chart-click': {
                    'dataSource': '{{ table_click_data }}',
                    'galleryId': '{{ table_gallery_id }}',                    
                    'getKey':(payload)=>payload.row[0],
                   },
                }"
              :command="{{ command }}"
              :data="{{ data }}"
            />"""

template_notification_str = """
            <div style="margin-top: 20px; margin-bottom: 20px;">
                <sly-iw-notification              
                iw-widget-id="{{ widget_id }}"
                :data="{{ data }}"
                >
                <span slot="title">
                    {{ title }}
                </span>

                <span slot="description">
                    {{ description }}
                </span>
                </sly-iw-notification>
            </div>"""


from collections import defaultdict
from typing import Callable, Optional, Literal

import numpy as np


def calculate_metrics(cocoGt, cocoDt, iouType: Literal["bbox", "segm"], progress_cb: Optional[Callable] = None):
    """
    Calculate COCO metrics.

    :param cocoGt: Ground truth dataset in COCO format
    :type cocoGt: COCO
    :param cocoDt: Predicted dataset in COCO format
    :type cocoDt: COCO
    :param iouType: Type of IoU calculation
    :type iouType: Literal["bbox", "segm"]
    :param progress_cb: Progress callback
    :type progress_cb: Optional[Callable]
    :return: Results of the evaluation
    :rtype: dict
    """
    from pycocotools.cocoeval import COCOeval  # pylint: disable=import-error

    progress_cb(1) if progress_cb is not None else None
    cocoEval = COCOeval(cocoGt, cocoDt, iouType=iouType)
    progress_cb(1) if progress_cb is not None else None
    cocoEval.evaluate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval.accumulate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval.summarize()
    progress_cb(1) if progress_cb is not None else None

    # For classification metrics
    cocoEval_cls = COCOeval(cocoGt, cocoDt, iouType=iouType)
    progress_cb(1) if progress_cb is not None else None
    cocoEval_cls.params.useCats = 0
    cocoEval_cls.evaluate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval_cls.accumulate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval_cls.summarize()
    progress_cb(1) if progress_cb is not None else None

    eval_img_dict = get_eval_img_dict(cocoEval)
    eval_img_dict_cls = get_eval_img_dict(cocoEval_cls)
    matches = get_matches(eval_img_dict, eval_img_dict_cls, cocoEval_cls, iou_t=0)

    params = {"iouThrs": cocoEval.params.iouThrs, "recThrs": cocoEval.params.recThrs}
    coco_metrics = {"mAP": cocoEval.stats[0], "precision": cocoEval.eval["precision"]}
    eval_data = {
        "matches": matches,
        "coco_metrics": coco_metrics,
        "params": params,
    }
    progress_cb(1) if progress_cb is not None else None

    return eval_data


def get_counts(cocoEval):
    """
    true_positives, false_positives, false_negatives

    type cocoEval: COCOeval
    """
    aRng = cocoEval.params.areaRng[0]
    cat_ids = cocoEval.params.catIds
    eval_imgs = [ev for ev in cocoEval.evalImgs if ev is not None and ev["aRng"] == aRng]

    N = len(eval_imgs)
    T = len(cocoEval.params.iouThrs)
    K = max(cat_ids) + 1

    true_positives = np.zeros((K, N, T))
    false_positives = np.zeros((K, N, T))
    false_negatives = np.zeros((K, N, T))

    for i, eval_img in enumerate(eval_imgs):
        catId = eval_img["category_id"]
        dt_matches = eval_img["dtMatches"]
        gt_matches = eval_img["gtMatches"]

        # Ignore
        if np.any(eval_img["gtIgnore"]):
            dt_matches = eval_img["dtMatches"].copy()
            dt_matches[eval_img["dtIgnore"]] = -1

            gt_matches = eval_img["gtMatches"].copy()
            gt_ignore_mask = eval_img["gtIgnore"][None,].repeat(T, axis=0).astype(bool)
            gt_matches[gt_ignore_mask] = -1

        true_positives[catId, i] = np.sum(dt_matches > 0, axis=1)
        false_positives[catId, i] = np.sum(dt_matches == 0, axis=1)
        false_negatives[catId, i] = np.sum(gt_matches == 0, axis=1)

    return true_positives[cat_ids], false_positives[cat_ids], false_negatives[cat_ids]


def get_counts_and_scores(cocoEval, cat_id: int, t: int):
    """
    tps, fps, scores, n_positives

    type cocoEval: COCOeval
    """
    aRng = cocoEval.params.areaRng[0]
    eval_imgs = [ev for ev in cocoEval.evalImgs if ev is not None and ev["aRng"] == aRng]

    tps = []
    fps = []
    # fns = []
    scores = []
    n_positives = 0

    # Process each evaluated image
    for eval_img in eval_imgs:
        if eval_img["category_id"] != cat_id:
            continue
        dtScores = eval_img["dtScores"]
        dtm = eval_img["dtMatches"][t]
        gtm = eval_img["gtMatches"][t]

        # ntp = (dtm > 0).sum()
        # nfp = (dtm == 0).sum()
        # nfn = (gtm == 0).sum()
        p = len(gtm)

        tp = (dtm > 0).astype(int).tolist()
        fp = (dtm == 0).astype(int).tolist()
        # fn = [nfn]*len(dtm)

        tps.extend(tp)
        fps.extend(fp)
        # fns.extend(fn)
        scores.extend(dtScores)
        n_positives += p

    assert len(tps) == len(fps) == len(scores)

    # sort by score
    indices = np.argsort(scores)[::-1]
    scores = np.array(scores)[indices]
    tps = np.array(tps)[indices]
    fps = np.array(fps)[indices]

    return tps, fps, scores, n_positives


def get_eval_img_dict(cocoEval):
    """
    type cocoEval: COCOeval
    """
    aRng = cocoEval.params.areaRng[0]
    eval_img_dict = defaultdict(list)  # img_id : dt/gt
    for i, eval_img in enumerate(cocoEval.evalImgs):
        if eval_img is None or eval_img["aRng"] != aRng:
            continue
        img_id = eval_img["image_id"]
        cat_id = eval_img["category_id"]
        ious = cocoEval.ious[(img_id, cat_id)]
        # ! inplace operation
        eval_img["ious"] = ious
        eval_img_dict[img_id].append(eval_img)
    eval_img_dict = dict(eval_img_dict)
    return eval_img_dict


def get_eval_img_dict_cls(cocoEval_cls):
    """
    type cocoEval_cls: COCOeval
    """
    # For miss-classification
    aRng = cocoEval_cls.params.areaRng[0]
    eval_img_dict_cls = defaultdict(list)  # img_id : dt/gt
    for i, eval_img in enumerate(cocoEval_cls.evalImgs):
        if eval_img is None or eval_img["aRng"] != aRng:
            continue
        img_id = eval_img["image_id"]
        cat_id = eval_img["category_id"]
        ious = cocoEval_cls.ious[(img_id, cat_id)]
        # ! inplace operation
        eval_img["ious"] = ious
        eval_img_dict_cls[img_id].append(eval_img)
    eval_img_dict_cls = dict(eval_img_dict_cls)
    assert np.all([len(x) == 1 for x in eval_img_dict_cls.values()])
    return eval_img_dict_cls


def _get_missclassified_match(eval_img_cls, dt_id, gtIds_orig, dtIds_orig, iou_t):
    # Correction on miss-classification
    gt_idx = np.nonzero(eval_img_cls["gtMatches"][iou_t] == dt_id)[0]
    if len(gt_idx) == 1:
        gt_idx = gt_idx[0]
        gt_id = eval_img_cls["gtIds"][gt_idx]
        gt_idx_o = gtIds_orig.index(gt_id)
        dt_idx_o = dtIds_orig.index(dt_id)
        iou = eval_img_cls["ious"][dt_idx_o, gt_idx_o].item()
        is_gt_ignore = eval_img_cls["gtIgnore"][gt_idx]
        if not is_gt_ignore:
            return gt_id, iou
    elif len(gt_idx) > 1:
        raise ValueError("Multiple matches")
    return None, None


def get_matches(eval_img_dict: dict, eval_img_dict_cls: dict, cocoEval_cls, iou_t: int = 0):
    """
    type cocoEval_cls: COCOeval
    """
    catIds = cocoEval_cls.cocoGt.getCatIds()
    matches = []
    for img_id, eval_imgs in eval_img_dict.items():

        # get miss-classified
        eval_img_cls = eval_img_dict_cls[img_id][0]
        gtIds_orig_cls = [_["id"] for cId in catIds for _ in cocoEval_cls._gts[img_id, cId]]

        for eval_img in eval_imgs:

            dtIds = np.array(eval_img["dtIds"])
            gtIds = np.array(eval_img["gtIds"])
            dtm = eval_img["dtMatches"][iou_t]
            gtm = eval_img["gtMatches"][iou_t]
            dtIgnore = eval_img["dtIgnore"][iou_t]
            gtIgnore = eval_img["gtIgnore"]

            # True Positives
            tp_idxs = np.nonzero(dtm)[0]
            for i in tp_idxs:
                if dtIgnore[i]:
                    continue
                dt_id = dtIds[i]
                gt_id = int(dtm[i])
                gt_idx = np.where(gtIds == gt_id)[0]
                iou = eval_img["ious"][i, gt_idx].item()
                score = eval_img["dtScores"][i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": dt_id,
                    "gt_id": gt_id,
                    "type": "TP",
                    "score": score,
                    "iou": iou,
                    "miss_cls": False,
                }
                assert iou >= 0.5, iou
                matches.append(match)

            # False Positives
            fp_idxs = np.nonzero(dtm == 0)[0]
            for i in fp_idxs:
                dt_id = dtIds[i]
                score = eval_img["dtScores"][i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": dt_id,
                    "gt_id": None,
                    "type": "FP",
                    "score": score,
                    "iou": None,
                    "miss_cls": False,
                }

                # Correction on miss-classification
                cls_gt_id, iou = _get_missclassified_match(
                    eval_img_cls, dt_id, gtIds_orig_cls, eval_img_cls["dtIds"], iou_t
                )
                if cls_gt_id is not None:
                    assert iou >= 0.5, iou
                    match["gt_id"] = cls_gt_id
                    match["iou"] = iou
                    match["miss_cls"] = True
                matches.append(match)

            # False Negatives
            fn_idxs = np.nonzero(gtm == 0)[0]
            for i in fn_idxs:
                if gtIgnore[i]:
                    continue
                gt_id = gtIds[i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": None,
                    "gt_id": gt_id,
                    "type": "FN",
                    "score": None,
                    "iou": None,
                    "miss_cls": False,
                }
                matches.append(match)

    return matches


def get_rare_classes(cocoGt, topk_ann_fraction=0.1, topk_classes_fraction=0.2):
    """
    :param cocoGt: Ground truth dataset in COCO format
    :type cocoGt: COCO
    """
    anns_cat_ids = [ann["category_id"] for ann in cocoGt.anns.values()]
    cat_ids, cat_counts = np.unique(anns_cat_ids, return_counts=True)
    inds_sorted = np.argsort(cat_counts)
    cum_counts = cat_counts[inds_sorted].cumsum()

    topk_upper1 = np.sum(cum_counts < cum_counts[-1] * topk_ann_fraction)
    topk_upper2 = int(topk_classes_fraction * len(cat_counts))
    topk_lower1 = 1

    topk = max(topk_lower1, min(topk_upper1, topk_upper2))
    cat_ids_rare = cat_ids[inds_sorted][:topk]
    cat_names = [cocoGt.cats[cat_id]["name"] for cat_id in cat_ids_rare]
    return cat_ids_rare, cat_names



from __future__ import annotations

from typing import TYPE_CHECKING

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class Precision(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.clickable = True
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_P=Widget.Markdown(title="Precision", is_header=True),
            notification_precision=Widget.Notification(
                formats_title=[self._loader.base_metrics()["precision"].round(2)],
                formats_desc=[
                    self._loader.mp.TP_count,
                    (self._loader.mp.TP_count + self._loader.mp.FP_count),
                ],
            ),
            markdown_P_perclass=Widget.Markdown(
                formats=[self._loader.vis_texts.definitions.f1_score]
            ),
            chart=Widget.Chart(),
        )

    def get_figure(self, widget: Widget):  #  -> Optional[go.Figure]
        import plotly.express as px  # pylint: disable=import-error

        # Per-class Precision bar chart
        # per_class_metrics_df_sorted = per_class_metrics_df.sort_values(by="precision")
        sorted_by_precision = self._loader.mp.per_class_metrics().sort_values(by="precision")
        fig = px.bar(
            sorted_by_precision,
            x="category",
            y="precision",
            # title="Per-class Precision (Sorted by F1)",
            color="precision",
            color_continuous_scale="Plasma",
        )
        fig.update_traces(hovertemplate="Class: %{x}<br>Precision: %{y:.2f}<extra></extra>")
        if len(sorted_by_precision) <= 20:
            fig.update_traces(
                text=sorted_by_precision.round(2),
                textposition="outside",
            )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="Precision", range=[0, 1])
        return fig



from __future__ import annotations

import json
import pickle
from typing import TYPE_CHECKING, Dict, List, Tuple

import pandas as pd
from jinja2 import Template

from supervisely import AnyGeometry, Bitmap, Polygon, Rectangle
from supervisely._utils import batched
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagApplicableTo, TagMeta, TagValueType
from supervisely.api.image_api import ImageInfo
from supervisely.convert.image.coco.coco_helper import HiddenCocoPrints
from supervisely.io import fs
from supervisely.io.fs import file_exists, mkdir
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta

if TYPE_CHECKING:
    from supervisely.nn.benchmark.base_benchmark import BaseBenchmark

from supervisely import Label
from supervisely.nn.benchmark.evaluation.coco.metric_provider import MetricProvider
from supervisely.nn.benchmark.visualization.vis_click_data import ClickData, IdMapper
from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_metrics import ALL_METRICS
from supervisely.nn.benchmark.visualization.vis_templates import generate_main_template
from supervisely.nn.benchmark.visualization.vis_widgets import Widget
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class ImageComparisonData:
    def __init__(
        self,
        gt_image_info: ImageInfo = None,
        pred_image_info: ImageInfo = None,
        diff_image_info: ImageInfo = None,
        gt_annotation: Annotation = None,
        pred_annotation: Annotation = None,
        diff_annotation: Annotation = None,
    ):
        self.gt_image_info = gt_image_info
        self.pred_image_info = pred_image_info
        self.diff_image_info = diff_image_info
        self.gt_annotation = gt_annotation
        self.pred_annotation = pred_annotation
        self.diff_annotation = diff_annotation


class Visualizer:

    def __init__(self, benchmark: BaseBenchmark) -> None:

        if benchmark.dt_project_info is None:
            raise RuntimeError(
                "The benchmark prediction project was not initialized. Please run evaluation or specify dt_project_info property of benchmark object."
            )

        eval_dir = benchmark.get_eval_results_dir()
        assert not fs.dir_empty(
            eval_dir
        ), f"The result dir {eval_dir!r} is empty. You should run evaluation before visualizing results."

        self._benchmark = benchmark
        self._api = benchmark.api
        self.cv_task = benchmark.cv_task

        self.eval_dir = benchmark.get_eval_results_dir()
        self.layout_dir = benchmark.get_layout_results_dir()

        self.dt_project_info = benchmark.dt_project_info
        self.gt_project_info = benchmark.gt_project_info
        self._benchmark.diff_project_info, existed = self._benchmark._get_or_create_diff_project()
        self.diff_project_info = benchmark.diff_project_info
        self.classes_whitelist = benchmark.classes_whitelist
        self.diff_project_meta = ProjectMeta.from_json(
            self._api.project.get_meta(self.diff_project_info.id)
        )
        self.comparison_data: Dict[int, ImageComparisonData] = {}  # gt_id -> ImageComparisonData

        self.gt_project_meta = self._get_filtered_project_meta(self.gt_project_info.id)
        self.dt_project_meta = self._get_filtered_project_meta(self.dt_project_info.id)
        self._docs_link = "https://docs.supervisely.com/neural-networks/model-evaluation-benchmark/"

        if benchmark.cv_task == CVTask.OBJECT_DETECTION:
            self._initialize_object_detection_loader()
            self.docs_link = self._docs_link + CVTask.OBJECT_DETECTION.value.replace("_", "-")
        elif benchmark.cv_task == CVTask.INSTANCE_SEGMENTATION:
            self._initialize_instance_segmentation_loader()
            self.docs_link = self._docs_link + CVTask.INSTANCE_SEGMENTATION.value.replace("_", "-")
        else:
            raise NotImplementedError(f"CV task {benchmark.cv_task} is not supported yet")

        self.pbar = benchmark.pbar

        if not existed:
            self.update_diff_annotations()
        else:
            self._init_comparison_data()

    def _initialize_object_detection_loader(self):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        from supervisely.nn.benchmark.visualization import vis_texts

        self.vis_texts = vis_texts

        cocoGt_path, cocoDt_path, eval_data_path, inference_info_path = (
            self.eval_dir + "/cocoGt.json",
            self.eval_dir + "/cocoDt.json",
            self.eval_dir + "/eval_data.pkl",
            self.eval_dir + "/inference_info.json",
        )

        with open(cocoGt_path, "r") as f:
            cocoGt_dataset = json.load(f)
        with open(cocoDt_path, "r") as f:
            cocoDt_dataset = json.load(f)

        # Remove COCO read logs
        with HiddenCocoPrints():
            cocoGt = COCO()
            cocoGt.dataset = cocoGt_dataset
            cocoGt.createIndex()
            cocoDt = cocoGt.loadRes(cocoDt_dataset["annotations"])

        with open(eval_data_path, "rb") as f:
            eval_data = pickle.load(f)

        inference_info = {}
        if file_exists(inference_info_path):
            with open(inference_info_path, "r") as f:
                inference_info = json.load(f)
            self.inference_info = inference_info
        else:
            self.inference_info = self._benchmark._eval_inference_info

        self.mp = MetricProvider(
            eval_data["matches"],
            eval_data["coco_metrics"],
            eval_data["params"],
            cocoGt,
            cocoDt,
        )
        self.mp.calculate()

        self.df_score_profile = pd.DataFrame(
            self.mp.confidence_score_profile(), columns=["scores", "precision", "recall", "f1"]
        )

        # downsample
        if len(self.df_score_profile) > 5000:
            self.dfsp_down = self.df_score_profile.iloc[:: len(self.df_score_profile) // 1000]
        else:
            self.dfsp_down = self.df_score_profile

        self.f1_optimal_conf = self.mp.get_f1_optimal_conf()[0]
        if self.f1_optimal_conf is None:
            self.f1_optimal_conf = 0.01
            logger.warn("F1 optimal confidence cannot be calculated. Using 0.01 as default.")

        # Click data
        gt_id_mapper = IdMapper(cocoGt_dataset)
        dt_id_mapper = IdMapper(cocoDt_dataset)

        self.click_data = ClickData(self.mp.m, gt_id_mapper, dt_id_mapper)
        self.base_metrics = self.mp.base_metrics

        self._objects_bindings = []

    def _initialize_instance_segmentation_loader(self):
        from pycocotools.coco import COCO  # pylint: disable=import-error

        from supervisely.nn.benchmark.visualization.instance_segmentation import (
            text_template,
        )

        self.vis_texts = text_template

        cocoGt_path, cocoDt_path, eval_data_path, inference_info_path = (
            self.eval_dir + "/cocoGt.json",
            self.eval_dir + "/cocoDt.json",
            self.eval_dir + "/eval_data.pkl",
            self.eval_dir + "/inference_info.json",
        )

        with open(cocoGt_path, "r") as f:
            cocoGt_dataset = json.load(f)
        with open(cocoDt_path, "r") as f:
            cocoDt_dataset = json.load(f)

        # Remove COCO read logs
        with HiddenCocoPrints():
            cocoGt = COCO()
            cocoGt.dataset = cocoGt_dataset
            cocoGt.createIndex()
            cocoDt = cocoGt.loadRes(cocoDt_dataset["annotations"])

        with open(eval_data_path, "rb") as f:
            eval_data = pickle.load(f)

        inference_info = {}
        if file_exists(inference_info_path):
            with open(inference_info_path, "r") as f:
                inference_info = json.load(f)
            self.inference_info = inference_info
        else:
            self.inference_info = self._benchmark._eval_inference_info

        self.mp = MetricProvider(
            eval_data["matches"],
            eval_data["coco_metrics"],
            eval_data["params"],
            cocoGt,
            cocoDt,
        )
        self.mp.calculate()

        self.df_score_profile = pd.DataFrame(
            self.mp.confidence_score_profile(), columns=["scores", "precision", "recall", "f1"]
        )

        # downsample
        if len(self.df_score_profile) > 5000:
            self.dfsp_down = self.df_score_profile.iloc[:: len(self.df_score_profile) // 1000]
        else:
            self.dfsp_down = self.df_score_profile

        self.f1_optimal_conf = self.mp.get_f1_optimal_conf()[0]
        if self.f1_optimal_conf is None:
            self.f1_optimal_conf = 0.01
            logger.warn("F1 optimal confidence cannot be calculated. Using 0.01 as default.")

        # Click data
        gt_id_mapper = IdMapper(cocoGt_dataset)
        dt_id_mapper = IdMapper(cocoDt_dataset)

        self.click_data = ClickData(self.mp.m, gt_id_mapper, dt_id_mapper)
        self.base_metrics = self.mp.base_metrics

        self._objects_bindings = []

    def visualize(self):
        from supervisely.app.widgets import GridGalleryV2

        mkdir(f"{self.layout_dir}/data", remove_content_if_exists=True)

        initialized = [mv(self) for mv in ALL_METRICS]
        initialized = [mv for mv in initialized if self.cv_task.value in mv.cv_tasks]
        with self.pbar(
            message="Saving visualization files",
            total=len([w for mv in initialized for w in mv.schema]),
        ) as p:
            for mv in initialized:
                for widget in mv.schema:
                    self._write_markdown_files(mv, widget)
                    self._write_json_files(mv, widget)
                    p.update(1)

        res = {}
        gallery = GridGalleryV2(
            columns_number=3,
            enable_zoom=False,
            annotations_opacity=0.4,
            border_width=4,
            default_tag_filters=[{"confidence": [self.f1_optimal_conf, 1]}],
            show_zoom_slider=False,
        )
        gallery._update_filters()
        res.update(gallery.get_json_state())

        self.dt_project_meta = self._get_filtered_project_meta(self.dt_project_info.id)
        res["projectMeta"] = self.dt_project_meta.to_json()
        for basename in ["modal_general.json", "modal_general_diff.json"]:
            local_path = f"{self.layout_dir}/data/{basename}"
            with open(local_path, "w", encoding="utf-8") as f:
                f.write(json.dumps(res))
            logger.info("Saved: %r", basename)

        self._save_template(initialized)

    def _write_markdown_files(self, metric_visualization: MetricVis, widget: Widget):

        if isinstance(widget, Widget.Markdown):
            content = metric_visualization.get_md_content(widget)
            local_path = f"{self.layout_dir}/data/{widget.name}.md"
            with open(local_path, "w", encoding="utf-8") as f:
                f.write(content)

            logger.info("Saved: %r", f"{widget.name}.md")

        if isinstance(widget, Widget.Collapse):
            for subwidget in widget.schema:
                content = metric_visualization.get_md_content(subwidget)
                local_path = f"{self.layout_dir}/data/{subwidget.name}.md"
                with open(local_path, "w", encoding="utf-8") as f:
                    f.write(content)

                logger.info("Saved: %r", f"{subwidget.name}.md")

    def _write_json_files(self, mv: MetricVis, widget: Widget):
        if isinstance(widget, Widget.Chart):
            fig = mv.get_figure(widget)
            if fig is not None:
                fig_data = {
                    "selected": None,
                    "galleryContent": "",
                    "dialogVisible": False,
                    "chartContent": json.loads(fig.to_json()),
                }
                basename = f"{widget.name}_{mv.name}.json"
                local_path = f"{self.layout_dir}/data/{basename}"
                with open(local_path, "w", encoding="utf-8") as f:
                    json.dump(fig_data, f)
                logger.info("Saved: %r", basename)

                click_data = mv.get_click_data(widget)
                if click_data is not None:
                    basename = f"{widget.name}_{mv.name}_click_data.json"
                    local_path = f"{self.layout_dir}/data/{basename}"
                    with open(local_path, "w", encoding="utf-8") as f:
                        f.write(json.dumps(click_data))
                    logger.info("Saved: %r", basename)

                    # modal_data = mv.get_modal_data(widget)
                    # basename = f"{widget.name}_{mv.name}_modal_data.json"
                    # local_path = f"{self.layout_dir}/data/{basename}"
                    # with open(local_path, "w", encoding="utf-8") as f:
                    #     f.write(json.dumps(modal_data))
                    # logger.info("Saved: %r", basename)

        if isinstance(widget, Widget.Gallery):
            content = mv.get_gallery(widget)
            if content is not None:
                basename = f"{widget.name}_{mv.name}.json"
                local_path = f"{self.layout_dir}/data/{basename}"
                with open(local_path, "w", encoding="utf-8") as f:
                    f.write(json.dumps(content))
                logger.info("Saved: %r", basename)

                click_data = mv.get_gallery_click_data(widget)
                if click_data is not None:
                    basename = f"{widget.name}_{mv.name}_click_data.json"
                    local_path = f"{self.layout_dir}/data/{basename}"
                    with open(local_path, "w", encoding="utf-8") as f:
                        f.write(json.dumps(click_data))
                    logger.info("Saved: %r", basename)

                diff_data = mv.get_diff_gallery_data(widget)
                if diff_data is not None:
                    basename = f"{widget.name}_{mv.name}_diff_data.json"
                    local_path = f"{self.layout_dir}/data/{basename}"
                    with open(local_path, "w", encoding="utf-8") as f:
                        f.write(json.dumps(diff_data))
                    logger.info("Saved: %r", basename)

                    # modal_data = mv.get_gallery_modal(widget)
                    # basename = f"{widget.name}_{mv.name}_modal_data.json"
                    # local_path = f"{self.layout_dir}/data/{basename}"
                    # with open(local_path, "w", encoding="utf-8") as f:
                    #     f.write(json.dumps(modal_data))
                    # logger.info("Saved: %r", basename)

        if isinstance(widget, Widget.Table):
            content = mv.get_table(widget)
            if content is not None:
                basename = f"{widget.name}_{mv.name}.json"
                local_path = f"{self.layout_dir}/data/{basename}"
                with open(local_path, "w", encoding="utf-8") as f:
                    f.write(json.dumps(content))
                logger.info("Saved: %r", basename)

                content = mv.get_table_click_data(widget)
                basename = f"{widget.name}_{mv.name}_click_data.json"
                local_path = f"{self.layout_dir}/data/{basename}"
                with open(local_path, "w", encoding="utf-8") as f:
                    f.write(json.dumps(content))
                logger.info("Saved: %r", basename)

    def _generate_template(self, metric_visualizations: Tuple[MetricVis]) -> str:
        html_snippets = {}
        main_template = Template(generate_main_template(metric_visualizations))
        for mv in metric_visualizations:
            for widget in mv.schema:
                if isinstance(widget, Widget.Notification):
                    html_snippets.update(mv.get_html_snippets())

            html_snippets.update(mv.get_html_snippets())

        return main_template.render(**html_snippets)

    def _generate_state(self, metric_visualizations: Tuple[MetricVis]) -> dict:
        res = {}
        for mv in metric_visualizations:
            for widget in mv.schema:
                if isinstance(widget, Widget.Chart) and mv.switchable:
                    res[mv.radiogroup_id] = widget.switch_key
                    break
        return res

    def _save_template(self, metric_visualizations: Tuple[MetricVis]):
        local_path = f"{self.layout_dir}/template.vue"
        with open(local_path, "w", encoding="utf-8") as f:
            f.write(self._generate_template(metric_visualizations))
        logger.info("Saved: %r", "template.vue")
        local_path = f"{self.layout_dir}/state.json"
        with open(local_path, "w", encoding="utf-8") as f:
            json.dump(self._generate_state(metric_visualizations), f)
        logger.info("Saved: %r", "state.json")

    def update_diff_annotations(self):
        meta = self._update_pred_meta_with_tags(self.dt_project_info.id, self.dt_project_meta)
        self._update_diff_meta(meta)

        self.dt_project_meta = meta
        self._add_tags_to_pred_project(self.mp.matches, self.dt_project_info.id)
        gt_project_path, pred_project_path = self._benchmark._download_projects(save_images=False)

        gt_project = Project(gt_project_path, OpenMode.READ)
        pred_project = Project(pred_project_path, OpenMode.READ)
        diff_dataset_name_to_info = {
            ds.name: ds for ds in self._api.dataset.get_list(self.diff_project_info.id)
        }

        matched_id_map = self._get_matched_id_map()  # dt_id -> gt_id
        matched_gt_ids = set(matched_id_map.values())

        outcome_tag = meta.get_tag_meta("outcome")
        conf_meta = meta.get_tag_meta("confidence")
        if conf_meta is None:
            conf_meta = meta.get_tag_meta("conf")
        match_tag = meta.get_tag_meta("matched_gt_id")

        pred_tag_list = []
        with self.pbar(message="Creating diff_project", total=pred_project.total_items) as progress:
            for pred_dataset in pred_project.datasets:
                pred_dataset: Dataset
                gt_dataset: Dataset = gt_project.datasets.get(pred_dataset.name)
                diff_dataset_info = diff_dataset_name_to_info[pred_dataset.name]
                diff_anns = []
                gt_image_ids = []
                pred_img_ids = []
                for item_name in pred_dataset.get_items_names():
                    gt_image_info = gt_dataset.get_image_info(item_name)
                    gt_image_ids.append(gt_image_info.id)
                    pred_image_info = pred_dataset.get_image_info(item_name)
                    pred_img_ids.append(pred_image_info.id)
                    gt_ann = gt_dataset.get_ann(item_name, gt_project.meta)
                    pred_ann = pred_dataset.get_ann(item_name, pred_project.meta)
                    labels = []

                    # TP and FP
                    for label in pred_ann.labels:
                        match_tag_id = matched_id_map.get(label.geometry.sly_id)
                        value = "TP" if match_tag_id else "FP"
                        pred_tag_list.append(
                            {
                                "tagId": outcome_tag.sly_id,
                                "figureId": label.geometry.sly_id,
                                "value": value,
                            }
                        )
                        conf = 1
                        for tag in label.tags.items():
                            tag: Tag
                            if tag.name in ["confidence", "conf"]:
                                conf = tag.value
                                break

                        if conf < self.f1_optimal_conf:
                            continue  # do not add labels with low confidence to diff project
                        if match_tag_id:
                            continue  # do not add TP labels to diff project
                        label = label.add_tag(Tag(outcome_tag, value))
                        label = label.add_tag(Tag(match_tag, int(label.geometry.sly_id)))
                        labels.append(label)

                    # FN
                    for label in gt_ann.labels:
                        if self.classes_whitelist:
                            if label.obj_class.name not in self.classes_whitelist:
                                continue
                        if label.geometry.sly_id not in matched_gt_ids:
                            if self._is_label_compatible_to_cv_task(label):
                                new_label = label.add_tags(
                                    [Tag(outcome_tag, "FN"), Tag(conf_meta, 1)]
                                )
                                labels.append(new_label)

                    diff_ann = Annotation(gt_ann.img_size, labels)
                    diff_anns.append(diff_ann)

                    # comparison data
                    self._update_comparison_data(
                        gt_image_info.id,
                        gt_image_info=gt_image_info,
                        pred_image_info=pred_image_info,
                        gt_annotation=gt_ann,
                        pred_annotation=pred_ann,
                        diff_annotation=diff_ann,
                    )

                diff_img_infos = self._api.image.copy_batch(diff_dataset_info.id, pred_img_ids)
                self._api.annotation.upload_anns(
                    [img_info.id for img_info in diff_img_infos],
                    diff_anns,
                    progress_cb=progress.update,
                )
                for gt_img_id, diff_img_info in zip(gt_image_ids, diff_img_infos):
                    self._update_comparison_data(gt_img_id, diff_image_info=diff_img_info)

        self._api.image.tag.add_to_objects(self.dt_project_info.id, pred_tag_list)

    def _init_comparison_data(self):
        gt_project_path, pred_project_path = self._benchmark._download_projects(save_images=False)
        gt_project = Project(gt_project_path, OpenMode.READ)
        pred_project = Project(pred_project_path, OpenMode.READ)
        diff_dataset_name_to_info = {
            ds.name: ds for ds in self._api.dataset.get_list(self.diff_project_info.id)
        }

        for pred_dataset in pred_project.datasets:
            pred_dataset: Dataset
            gt_dataset: Dataset = gt_project.datasets.get(pred_dataset.name)
            try:
                diff_dataset_info = diff_dataset_name_to_info[pred_dataset.name]
            except KeyError:
                raise RuntimeError(
                    f"Difference project was not created properly. Dataset {pred_dataset.name} is missing"
                )

            for item_names_batch in batched(pred_dataset.get_items_names(), 100):
                # diff project may be not created yet
                item_names_batch.sort()
                try:
                    diff_img_infos_batch: List[ImageInfo] = sorted(
                        self._api.image.get_list(
                            diff_dataset_info.id,
                            filters=[
                                {"field": "name", "operator": "in", "value": item_names_batch}
                            ],
                        ),
                        key=lambda x: x.name,
                    )
                    diff_anns_batch_dict = {
                        ann_info.image_id: Annotation.from_json(
                            ann_info.annotation, self.diff_project_meta
                        )
                        for ann_info in self._api.annotation.download_batch(
                            diff_dataset_info.id, [img_info.id for img_info in diff_img_infos_batch]
                        )
                    }
                    assert (
                        len(item_names_batch)
                        == len(diff_img_infos_batch)
                        == len(diff_anns_batch_dict)
                    ), "Some images are missing in the difference project"

                    for item_name, diff_img_info in zip(item_names_batch, diff_img_infos_batch):
                        assert (
                            item_name == diff_img_info.name
                        ), "Image names in difference project and prediction project do not match"
                        gt_image_info = gt_dataset.get_image_info(item_name)
                        pred_image_info = pred_dataset.get_image_info(item_name)
                        gt_ann = gt_dataset.get_ann(item_name, gt_project.meta)
                        pred_ann = pred_dataset.get_ann(item_name, pred_project.meta)
                        diff_ann = diff_anns_batch_dict[diff_img_info.id]

                        self._update_comparison_data(
                            gt_image_info.id,
                            gt_image_info=gt_image_info,
                            pred_image_info=pred_image_info,
                            diff_image_info=diff_img_info,
                            gt_annotation=gt_ann,
                            pred_annotation=pred_ann,
                            diff_annotation=diff_ann,
                        )
                except Exception:
                    raise RuntimeError("Difference project was not created properly")

    def _update_comparison_data(
        self,
        gt_image_id: int,
        gt_image_info: ImageInfo = None,
        pred_image_info: ImageInfo = None,
        diff_image_info: ImageInfo = None,
        gt_annotation: Annotation = None,
        pred_annotation: Annotation = None,
        diff_annotation: Annotation = None,
    ):
        comparison_data = self.comparison_data.get(gt_image_id, None)
        if comparison_data is None:
            self.comparison_data[gt_image_id] = ImageComparisonData(
                gt_image_info=gt_image_info,
                pred_image_info=pred_image_info,
                diff_image_info=diff_image_info,
                gt_annotation=gt_annotation,
                pred_annotation=pred_annotation,
                diff_annotation=diff_annotation,
            )
        else:
            for attr, value in {
                "gt_image_info": gt_image_info,
                "pred_image_info": pred_image_info,
                "diff_image_info": diff_image_info,
                "gt_annotation": gt_annotation,
                "pred_annotation": pred_annotation,
                "diff_annotation": diff_annotation,
            }.items():
                if value is not None:
                    setattr(comparison_data, attr, value)

    def _update_pred_meta_with_tags(self, project_id: int, meta: ProjectMeta) -> ProjectMeta:
        old_meta = meta
        outcome_tag = TagMeta(
            "outcome",
            value_type=TagValueType.ONEOF_STRING,
            possible_values=["TP", "FP", "FN"],
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        match_tag = TagMeta(
            "matched_gt_id",
            TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        iou_tag = TagMeta(
            "iou",
            TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        confidence_tag = TagMeta(
            "confidence",
            value_type=TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )

        for tag in [outcome_tag, match_tag, iou_tag]:
            if meta.get_tag_meta(tag.name) is None:
                meta = meta.add_tag_meta(tag)

        if meta.get_tag_meta("confidence") is None and meta.get_tag_meta("conf") is None:
            meta = meta.add_tag_meta(confidence_tag)

        if old_meta == meta:
            return meta

        meta = self._api.project.update_meta(project_id, meta)
        return meta

    def _update_diff_meta(self, meta: ProjectMeta):
        new_obj_classes = []
        for obj_class in meta.obj_classes:
            new_obj_classes.append(obj_class.clone(geometry_type=AnyGeometry))
        meta = meta.clone(obj_classes=new_obj_classes)
        self.diff_project_meta = self._api.project.update_meta(self.diff_project_info.id, meta)

    def _update_diff_meta(self, meta: ProjectMeta):
        new_obj_classes = []
        for obj_class in meta.obj_classes:
            new_obj_classes.append(obj_class.clone(geometry_type=AnyGeometry))
        meta = meta.clone(obj_classes=new_obj_classes)
        self.diff_project_meta = self._api.project.update_meta(self.diff_project_info.id, meta)

    def _add_tags_to_pred_project(self, matches: list, pred_project_id: int):

        # get tag metas
        # outcome_tag_meta = self.dt_project_meta.get_tag_meta("outcome")
        match_tag_meta = self.dt_project_meta.get_tag_meta("matched_gt_id")
        iou_tag_meta = self.dt_project_meta.get_tag_meta("iou")

        # mappings
        gt_ann_mapping = self.click_data.gt_id_mapper.map_obj
        dt_ann_mapping = self.click_data.dt_id_mapper.map_obj

        # add tags to objects
        logger.info("Adding tags to DT project")

        with self.pbar(message="Adding tags to DT project", total=len(matches)) as p:
            for batch in batched(matches, 100):
                pred_tag_list = []
                for match in batch:
                    if match["type"] == "TP":
                        outcome = "TP"
                        matched_gt_id = gt_ann_mapping[match["gt_id"]]
                        ann_dt_id = dt_ann_mapping[match["dt_id"]]
                        iou = match["iou"]
                        # api.advanced.add_tag_to_object(outcome_tag_meta.sly_id, ann_dt_id, str(outcome))
                        if matched_gt_id is not None:
                            pred_tag_list.extend(
                                [
                                    {
                                        "tagId": match_tag_meta.sly_id,
                                        "figureId": ann_dt_id,
                                        "value": int(matched_gt_id),
                                    },
                                    {
                                        "tagId": iou_tag_meta.sly_id,
                                        "figureId": ann_dt_id,
                                        "value": float(iou),
                                    },
                                ]
                            )
                        else:
                            continue
                    elif match["type"] == "FP":
                        outcome = "FP"
                        # api.advanced.add_tag_to_object(outcome_tag_meta.sly_id, ann_dt_id, str(outcome))
                    elif match["type"] == "FN":
                        outcome = "FN"
                    else:
                        raise ValueError(f"Unknown match type: {match['type']}")

                self._api.image.tag.add_to_objects(pred_project_id, pred_tag_list)
                p.update(len(batch))

    def _get_matched_id_map(self):
        gt_ann_mapping = self.click_data.gt_id_mapper.map_obj
        dt_ann_mapping = self.click_data.dt_id_mapper.map_obj
        dtId2matched_gt_id = {}
        for match in self.mp.matches_filtered:
            if match["type"] == "TP":
                dtId2matched_gt_id[dt_ann_mapping[match["dt_id"]]] = gt_ann_mapping[match["gt_id"]]
        return dtId2matched_gt_id

    def _is_label_compatible_to_cv_task(self, label: Label):
        if self.cv_task == CVTask.OBJECT_DETECTION:
            return isinstance(label.geometry, Rectangle)
        if self.cv_task == CVTask.INSTANCE_SEGMENTATION:
            return isinstance(label.geometry, (Bitmap, Polygon))
        return False

    def _get_filtered_project_meta(self, project_id: int) -> ProjectMeta:
        meta = self._api.project.get_meta(project_id)
        meta = ProjectMeta.from_json(meta)
        remove_classes = []
        if self.classes_whitelist:
            for obj_class in meta.obj_classes:
                if obj_class.name not in self.classes_whitelist:
                    remove_classes.append(obj_class.name)
            if remove_classes:
                meta = meta.delete_obj_classes(remove_classes)
        return meta



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer

from supervisely.project.project_meta import ProjectMeta


class ExplorerGrid(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.clickable = True
        self.has_diffs_view = True

        filters = [{"confidence": [self.f1_optimal_conf, 1]}]
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_explorer=Widget.Markdown(title="Explore Predictions", is_header=True),
            gallery=Widget.Gallery(filters=filters),
        )

    def _get_gallery(self, widget: Widget.Gallery, limit: Optional[int] = None) -> dict:
        res = {}
        api = self._loader._api
        pred_project_id = self._loader.dt_project_info.id
        pred_dataset = api.dataset.get_list(pred_project_id)[0]
        project_meta = ProjectMeta.from_json(api.project.get_meta(pred_project_id))
        pred_image_infos = api.image.get_list(dataset_id=pred_dataset.id, limit=limit)
        pred_image_ids = [x.id for x in pred_image_infos]
        ann_infos = api.annotation.download_batch(pred_dataset.id, pred_image_ids)

        for idx, (pred_image, ann_info) in enumerate(zip(pred_image_infos, ann_infos)):
            image_name = pred_image.name
            image_url = pred_image.full_storage_url
            widget.gallery.append(
                title=image_name,
                image_url=image_url,
                annotation_info=ann_info,
                column_index=idx % 3,
                project_meta=project_meta,
                ignore_tags_filtering=["outcome"],
            )
        res.update(widget.gallery.get_json_state())
        res.update(widget.gallery.get_json_data()["content"])
        res["layoutData"] = res.pop("annotations")
        res["projectMeta"] = project_meta.to_json()

        return res

    def get_gallery(self, widget: Widget.Gallery):
        return self._get_gallery(widget, limit=9)

    def get_gallery_click_data(self, widget: Widget.Gallery):
        res = {}

        res["layoutTemplate"] = [{"skipObjectTagsFiltering": ["outcome"]}] * 3
        click_data = res.setdefault("clickData", {})
        explore = click_data.setdefault("explore", {})
        explore["title"] = "Explore all predictions"
        images_ids = explore.setdefault("imagesIds", [])

        images_ids.extend([cd.pred_image_info.id for cd in self._loader.comparison_data.values()])

        return res

    def get_diff_gallery_data(self, widget: Widget.Gallery) -> Optional[dict]:
        res = {}

        res["layoutTemplate"] = [
            {"skipObjectTagsFiltering": True, "columnTitle": "Ground Truth"},
            {"skipObjectTagsFiltering": ["outcome"], "columnTitle": "Prediction"},
            {"skipObjectTagsFiltering": ["confidence"], "columnTitle": "Difference"},
        ]

        click_data = res.setdefault("clickData", {})

        default_filters = [
            {"type": "tag", "tagId": "confidence", "value": [self.f1_optimal_conf, 1]},
            # {"type": "tag", "tagId": "outcome", "value": "FP"},
        ]
        for img_comparison_data in self._loader.comparison_data.values():
            gt = img_comparison_data.gt_image_info
            pred = img_comparison_data.pred_image_info
            diff = img_comparison_data.diff_image_info
            assert gt.name == pred.name == diff.name
            key = click_data.setdefault(str(pred.id), {})
            key["imagesIds"] = [gt.id, pred.id, diff.id]
            key["filters"] = default_filters
            key["title"] = f"Image: {pred.name}"
            image_id = pred.id
            ann_json = img_comparison_data.pred_annotation.to_json()
            assert image_id == pred.id
            object_bindings = []
            for obj in ann_json["objects"]:
                for tag in obj["tags"]:
                    if tag["name"] == "matched_gt_id":
                        object_bindings.append(
                            [
                                {
                                    "id": obj["id"],
                                    "annotationKey": image_id,
                                },
                                {
                                    "id": int(tag["value"]),
                                    "annotationKey": gt.id,
                                },
                            ]
                        )

            image_id = diff.id
            ann_json = img_comparison_data.diff_annotation.to_json()
            assert image_id == diff.id
            for obj in ann_json["objects"]:
                for tag in obj["tags"]:
                    if tag["name"] == "matched_gt_id":
                        object_bindings.append(
                            [
                                {
                                    "id": obj["id"],
                                    "annotationKey": image_id,
                                },
                                {
                                    "id": int(tag["value"]),
                                    "annotationKey": pred.id,
                                },
                            ]
                        )
            key["objectsBindings"] = object_bindings

        return res

    # def get_gallery_modal(self, widget: Widget.Gallery):
    #     res = self.get_gallery(widget)

    #     res.pop("layout")
    #     res.pop("layoutData")

    #     return res



from typing import Union, List
from supervisely.app import DataJson
from supervisely.app.widgets import (
    Widget,
    TagMetaView,
    Checkbox,
    Grid,
    generate_id,
    Button,
)
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection


class TagMetasList(Widget):
    def __init__(
        self,
        tag_metas: Union[TagMetaCollection, List[TagMeta]],
        show_type_text: bool = True,
        limit_long_names: bool = False,
        selectable: bool = False,
        columns: int = 1,  # 1 means vertical layout
        widget_id: str = None,
    ):
        self._tag_metas = tag_metas
        self._selectable = selectable
        self._columns = columns

        if type(tag_metas) is list:
            self._tag_metas = TagMetaCollection(self._tag_metas)

        self._name_to_tag = {}
        self._name_to_view = {}
        self._name_to_checkbox = {}
        self._select_all_btn = None
        self._deselect_all_btn = None

        for tag_meta in self._tag_metas:
            self._name_to_tag[tag_meta.name] = tag_meta
            self._name_to_view[tag_meta.name] = TagMetaView(
                tag_meta, show_type_text, limit_long_names, widget_id=generate_id()
            )
            grid_items = list(self._name_to_view.values())
            if self._selectable is True:
                self._select_all_btn = Button(
                    "Select all",
                    button_type="text",
                    show_loading=False,
                    icon="zmdi zmdi-check-all",
                    widget_id=generate_id(),
                )
                self._deselect_all_btn = Button(
                    "Deselect all",
                    button_type="text",
                    show_loading=False,
                    icon="zmdi zmdi-square-o",
                    widget_id=generate_id(),
                )
                current_checkbox = Checkbox(
                    self._name_to_view[tag_meta.name],
                    checked=True,
                    widget_id=generate_id(),
                )
                self._name_to_checkbox[tag_meta.name] = current_checkbox
                grid_items = list(self._name_to_checkbox.values())

                @self._select_all_btn.click
                def select_all():
                    for k, v in self._name_to_checkbox.items():
                        v: Checkbox
                        v.check()
                    DataJson()[self.widget_id]["num_selected"] = len(self._name_to_checkbox)
                    DataJson().send_changes()

                @self._deselect_all_btn.click
                def deselect_all():
                    for k, v in self._name_to_checkbox.items():
                        v: Checkbox
                        v.uncheck()
                    DataJson()[self.widget_id]["num_selected"] = 0
                    DataJson().send_changes()

                @current_checkbox.value_changed
                def checkbox_changed(is_checked):
                    DataJson()[self.widget_id]["num_selected"] = len(self.get_selected_tag_names())
                    DataJson().send_changes()

        self._content = Grid(
            widgets=grid_items,
            columns=self._columns,
            widget_id=generate_id(),
        )
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        res = {"selectable": self._selectable}
        if self._selectable is True:
            res["num_selected"] = len(self.get_selected_tag_names())
        return res

    def get_selected_tag_names(self):
        if self._selectable is False:
            raise ValueError(
                "Class selection is disabled, because TagMetasList widget was created with argument 'selectable == False'"
            )
        results = []
        for k, v in self._name_to_checkbox.items():
            v: Checkbox
            if v.is_checked():
                results.append(k)
        return results

    def get_json_state(self):
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Schema, Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class ConfidenceScore(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.schema = Schema(
            self._loader.vis_texts,
            markdown_confidence_score_1=Widget.Markdown(
                title="Confidence Score Profile",
                is_header=True,
                formats=[self._loader.vis_texts.definitions.confidence_threshold],
            ),
            notification_f1=Widget.Notification(
                formats_title=[round((self._loader.mp.m_full.get_f1_optimal_conf()[0] or 0.0), 4)]
            ),
            chart=Widget.Chart(),
            markdown_confidence_score_2=Widget.Markdown(),
            collapse_conf_score=Widget.Collapse(
                Schema(
                    self._loader.vis_texts,
                    markdown_plot_confidence_profile=Widget.Markdown(
                        title="How to plot Confidence Profile?"
                    ),
                )
            ),
            markdown_calibration_score_3=Widget.Markdown(),
        )

    def get_figure(self, widget: Widget):  # -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error

        color_map = {
            "Precision": "#1f77b4",
            "Recall": "orange",
        }

        fig = px.line(
            self._loader.dfsp_down,
            x="scores",
            y=["precision", "recall", "f1"],
            # title="Confidence Score Profile",
            labels={"value": "Value", "variable": "Metric", "scores": "Confidence Score"},
            width=None,
            height=500,
            color_discrete_map=color_map,
        )
        fig.update_traces(
            hovertemplate="Confidence Score: %{x:.2f}<br>Value: %{y:.2f}<extra></extra>"
        )
        fig.update_layout(yaxis=dict(range=[0, 1]), xaxis=dict(range=[0, 1], tick0=0, dtick=0.1))

        if self._loader.mp.f1_optimal_conf is not None and self._loader.mp.best_f1 is not None:
            # Add vertical line for the best threshold
            fig.add_shape(
                type="line",
                x0=self._loader.mp.f1_optimal_conf,
                x1=self._loader.mp.f1_optimal_conf,
                y0=0,
                y1=self._loader.mp.best_f1,
                line=dict(color="gray", width=2, dash="dash"),
            )
            fig.add_annotation(
                x=self._loader.mp.f1_optimal_conf,
                y=self._loader.mp.best_f1 + 0.04,
                text=f"F1-optimal threshold: {self._loader.mp.f1_optimal_conf:.2f}",
                showarrow=False,
            )
        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig



from functools import wraps
from typing import Dict, List, Union

from supervisely._utils import batched, rand_str
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.empty.empty import Empty
from supervisely.app.widgets.line_chart.line_chart import LineChart
from supervisely.app.widgets.line_plot.line_plot import LinePlot
from supervisely.app.widgets.widget import Widget, generate_id
from supervisely.sly_logger import logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class GridChart(Widget):
    def __init__(
        self,
        data: List[Union[dict, str]] = [],
        columns: int = 1,
        gap: int = 10,
        widget_id: str = None,
    ):
        # TODO maybe generalize LineChart -> ApexChart ???
        self._widgets = {}
        self._columns = columns
        self._gap = gap

        if len(data) > 0:  # TODO: add the self.add_data([data:dict]) method
            for plot_data in data:
                if isinstance(plot_data, dict):
                    self._widgets[plot_data["title"]] = LineChart(**plot_data)
                else:
                    self._widgets[plot_data] = LineChart(title=plot_data, series=[])

            if self._columns < 1:
                raise ValueError(f"columns ({self._columns}) < 1")
            elif self._columns > len(self._widgets):
                logger.warn(
                    f"Number of columns ({self._columns}) > number of widgets ({len(self._widgets)}). Columns are set to {len(self._widgets)}"
                )
                self._columns = len(self._widgets)

            self._content = None
            if self._columns == 1:
                self._content = Container(
                    direction="vertical",
                    widgets=self._widgets.values(),
                    gap=self._gap,
                    widget_id=generate_id(),
                )
            else:
                rows = []
                num_empty = len(self._widgets) % self._columns
                for i in range(num_empty):
                    self._widgets[generate_id()] = Empty()
                for batch in batched(list(self._widgets.values()), batch_size=self._columns):
                    rows.append(
                        Container(
                            direction="horizontal",
                            widgets=batch,
                            gap=self._gap,
                            fractions=[1] * len(batch),
                            widget_id=generate_id(),
                            overflow=None,
                            widgets_style="flex-direction: column",
                        )
                    )
                self._content = Container(
                    direction="vertical",
                    widgets=rows,
                    gap=self._gap,
                    widget_id=generate_id(),
                )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None

    def add_scalar(
        self,
        identifier: str,
        y: Union[float, int],
        x: Union[float, int],
    ) -> None:
        """
        Add scalar to a single series on plot. If no series with name,
         defined in `identifier` exists, one will be created automatically.

        :param identifier: '/'-separated plot and series name
        :type identifier: str
        :param y: y value
        :type y: float | int
        :param x: x value
        :type x: float | int
        """
        plot_title, series_name = identifier.split("/")  # TODO: maybe Dict[str,str] is way better??
        w: LineChart = self._widgets[plot_title]
        series_id, series = w.get_series_by_name(series_name)

        if series is not None:
            w.add_to_series(name_or_id=series_id, data=[(x, y)])
        else:
            w.add_series(name=series_name, x=[x], y=[y])

    def add_scalars(
        self,
        plot_title: str,
        new_values: Dict[str, Union[float, int]],
        x: Union[float, int],
    ):
        """
        Add scalars to several series on one plot at point `x`

        :param plot_title: name of existing plot
        :type plot_title: str
        :param new_values: A dictionary in the `{'series_1': y1, 'series_2': y2, ...}` format.
        :type new_values: Dict[str, (float | int)]]
        :param x: value of `x`
        :type x: float | int
        """
        w: LineChart = self._widgets[plot_title]
        for s_name in new_values.keys():
            series_id, series = w.get_series_by_name(s_name)
            if series is not None:
                w.add_to_series(
                    name_or_id=series_id,
                    data=w._wrap_xy(x, new_values[s_name]),
                )
            else:
                w.add_series(name=s_name, x=[x], y=[new_values[s_name]])



from __future__ import annotations
from typing import Optional, List, Dict

from supervisely._utils import take_with_default
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import PointcloudEpisodeObjectCollection

class PointcloudEpisodeFrame(Frame):
    """
    PointcloudEpisodeFrame object for :class:`PointcloudEpisodeAnnotation<supervisely.pointcloud_annotation.pointcloud_episode_annotation.PointcloudEpisodeAnnotation>`. :class:`PointcloudEpisodeFrame<PointcloudEpisodeFrame>` object is immutable.

    :param index: Index of the PointcloudEpisodeFrame.
    :type index: int
    :param figures: List of :class:`PointcloudFigure<supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure>`.
    :type figures: list, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d
        from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import PointcloudEpisodeObjectCollection

        # Create pointcloud object
        obj_class_car = sly.ObjClass('car', Cuboid3d)
        pointcloud_obj_car = sly.PointcloudEpisodeObject(obj_class_car)
        objects = PointcloudEpisodeObjectCollection([pointcloud_obj_car])

        # Create figure
        frame_index = 7
        position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
        cuboid = Cuboid3d(position, rotation, dimension)

        figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

        frame = sly.PointcloudEpisodeFrame(frame_index, figures=[figure])
        print(frame.to_json())
        # Output: {
        #     "figures": [
        #         {
        #         "geometry": {
        #             "dimensions": { "x": 1.8, "y": 3.9, "z": 1.6 },
        #             "position": { "x": -3.4, "y": 28.9, "z": -0.7 },
        #             "rotation": { "x": 0.0, "y": 0, "z": -0.03 }
        #         },
        #         "geometryType": "cuboid_3d",
        #         "key": "cd61788d6faa401b9723f066f92a8a30",
        #         "objectKey": "c081cb9f34e54ff2bd85e04e7713ed76"
        #         }
        #     ],
        #     "index": 7
        # }
    """

    figure_type = PointcloudFigure

    def __init__(self, index: int, figures: Optional[List[PointcloudFigure]]=None):
        super().__init__(index, figures)

    @classmethod
    def from_json(
        cls, 
        data: Dict, 
        objects: PointcloudEpisodeObjectCollection, 
        frames_count: Optional[int]=None, 
        key_id_map: Optional[KeyIdMap]=None
    ) -> PointcloudEpisodeFrame:
        """
        Convert a json dict to PointcloudEpisodeFrame. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param objects: PointcloudEpisodeObjectCollection object.
        :type objects: PointcloudEpisodeObjectCollection
        :param frames_count: Number of frames in point cloud.
        :type frames_count: int, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :raises: :class:`ValueError` if frame index < 0 and if frame index > number of frames in point cloud
        :return: PointcloudEpisodeFrame object
        :rtype: :class:`PointcloudEpisodeFrame`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d
            from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import PointcloudEpisodeObjectCollection

            obj_class_car = sly.ObjClass('car', Cuboid3d)
            pointcloud_obj_car = sly.PointcloudEpisodeObject(obj_class_car)
            objects = PointcloudEpisodeObjectCollection([pointcloud_obj_car])

            frame_index = 7
            position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
            cuboid = Cuboid3d(position, rotation, dimension)

            figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

            frame = sly.PointcloudEpisodeFrame(frame_index, figures=[figure])
            frame_json = frame.to_json()

            frame_from_json = sly.PointcloudEpisodeFrame.from_json(frame_json, objects)
        """

        return super().from_json(data, objects, frames_count, key_id_map)

    def clone(self, index: Optional[int] = None, figures: Optional[List[PointcloudFigure]] = None) -> PointcloudEpisodeFrame:
        return super().clone(index, figures)


from typing import List

from supervisely.geometry.helpers import GET_GEOMETRY_FROM_STR
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.graph import GraphNodes, KeypointsTemplate, NODES
from supervisely.annotation.label import LabelJsonFields
from supervisely.annotation.tag import TagJsonFields
from supervisely import ObjClass, ProjectMeta, TagMeta, TagValueType, logger

SLY_IMAGE_ANN_KEYS = ["objects", "tags", "size"]
SLY_OBJECT_KEYS = [LabelJsonFields.OBJ_CLASS_NAME, LabelJsonFields.TAGS, "geometryType"]  #, LabelJsonFields.GEOMETRY_TYPE] TODO: add geometry type
SLY_TAG_KEYS = [TagJsonFields.TAG_NAME, TagJsonFields.VALUE]

# Check the annotation format documentation at
def get_meta_from_annotation(ann_json: dict, meta: ProjectMeta) -> ProjectMeta:
    """Generate sly.ProjectMeta from JSON annotation file."""

    if "annotation" in ann_json:
        ann_json = ann_json["annotation"]

    if not all(key in ann_json for key in SLY_IMAGE_ANN_KEYS):
        logger.warn(
            f"Some keys are missing in the annotation file. "
            "Check the annotation format documentation at: "
            "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/05_supervisely_format_images"
        )
        return meta

    for object in ann_json["objects"]:
        meta = create_tags_from_annotation(object["tags"], meta)
        meta = create_classes_from_annotation(object, meta)
    meta = create_tags_from_annotation(ann_json["tags"], meta)
    return meta


def create_tags_from_annotation(tags: List[dict], meta: ProjectMeta) -> ProjectMeta:
    for tag in tags:
        if not all(key in tag for key in SLY_TAG_KEYS):
            logger.warn(
                f"Tag in annotation file is not in Supervisely format. "
                "Read more about the Supervisely JSON format of tags in the documentation at: "
                "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/03_supervisely_format_tags"
            )
            continue
        tag_name = tag[TagJsonFields.TAG_NAME]
        tag_value = tag[TagJsonFields.VALUE]
        if tag_value is None:
            tag_meta = TagMeta(tag_name, TagValueType.NONE)
        elif isinstance(tag_value, int) or isinstance(tag_value, float):
            tag_meta = TagMeta(tag_name, TagValueType.ANY_NUMBER)
        else:
            tag_meta = TagMeta(tag_name, TagValueType.ANY_STRING)

        # check existing tag_meta in meta
        existing_tag = meta.get_tag_meta(tag_name)
        if existing_tag is None:
            meta = meta.add_tag_meta(tag_meta)
    return meta


def create_classes_from_annotation(object: dict, meta: ProjectMeta) -> ProjectMeta:
    if not all(key in object for key in SLY_OBJECT_KEYS):
        logger.warn(
            f"Object in annotation file is not in Supervisely format: {object}. "
            "Read more about the Supervisely JSON format of objects in the documentation at: "
            "https://docs.supervisely.com/customization-and-integration/00_ann_format_navi/04_supervisely_format_objects"
        )
        return meta
    class_name = object["classTitle"]
    geometry_type_str = object["geometryType"]
    obj_class = None

    try:
        geometry_type = GET_GEOMETRY_FROM_STR(geometry_type_str)
    except KeyError:
        logger.warn(f"Unknown geometry type {geometry_type_str} for class {class_name}")
        return meta

    obj_class = None
    geometry_config = None
    if issubclass(geometry_type, GraphNodes):
        if NODES in object:
            geometry_config = KeypointsTemplate()
            for uuid, node in object[NODES].items():
                if "loc" in node and len(node["loc"]) == 2:
                    geometry_config.add_point(label=uuid, row=node["loc"][0], col=node["loc"][1])
    obj_class = ObjClass(name=class_name, geometry_type=geometry_type, geometry_config=geometry_config)
    existing_class = meta.get_obj_class(class_name)

    if obj_class is None:
        logger.warn(f"Failed to create object class for {class_name} with geometry type {geometry_type_str}")
        return meta

    if existing_class is None:
        meta = meta.add_obj_class(obj_class)
    else:
        if existing_class.geometry_type != obj_class.geometry_type:
            obj_class = ObjClass(name=class_name, geometry_type=AnyGeometry)
            meta = meta.delete_obj_class(class_name)
            meta = meta.add_obj_class(obj_class)
    return meta

def rename_in_json(ann_json, renamed_classes=None, renamed_tags=None):
    if renamed_classes:
        for obj in ann_json["objects"]:
            obj["classTitle"] = renamed_classes.get(obj["classTitle"], obj["classTitle"])
            for tag in obj["tags"]:
                tag["name"] = renamed_tags.get(tag["name"], tag["name"])
    if renamed_tags:
        for tag in ann_json["tags"]:
            tag["name"] = renamed_tags.get(tag["name"], tag["name"])
    return ann_json

def annotation_high_level_validator(ann_json: dict) -> bool:
    """Check if annotation is probably in Supervisely format."""

    if "annotation" in ann_json:
        ann_json = ann_json["annotation"]
    if not all(key in ann_json for key in SLY_IMAGE_ANN_KEYS):
        return False
    for obj in ann_json["objects"]:
        if not all(key in obj for key in SLY_OBJECT_KEYS):
            return False
        for tag in obj["tags"]:
            if not all(key in tag for key in SLY_TAG_KEYS):
                return False
    for tag in ann_json["tags"]:
        if not all(key in tag for key in SLY_TAG_KEYS):
            return False
    return True

def get_image_size_from_annotation(ann_json: dict) -> tuple:
    """Get image size from annotation."""

    if "annotation" in ann_json:
        ann_json = ann_json["annotation"]
    if "size" not in ann_json:
        return None
    size = ann_json["size"]
    if "height" not in size or "width" not in size:
        return None
    return size["height"], size["width"]
